/*
 * Copyright (c) 2014 Patrick P. Frey
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
#ifndef _STRUS_BINDING_SWIG_HPP_INCLUDED
#define _STRUS_BINDING_SWIG_HPP_INCLUDED

#if defined DOXYGEN_JAVA
namespace java {
namespace lang {
class String
	:public std::string
{
public:
	///\brief Constructor
	String( char[]);
	///\brief Copy constructor
	String( const String& o);
};
}
namespace util {

/// \brief Interface of java abstract list as used by derived classes of strus
template <Element>
class AbstractList :public std::vector<Element>
{
public:
	///\brief Get an element by index
	///\param[in] i index starting from 0
	///\return the element reference
	const_reference get(int i) const throw (std::out_of_range);
	///\brief Set an element
	///\param[in] i index starting from 0
	///\param[in] val value of the element to set
	value_type set(int i, const value_type& val) throw (std::out_of_range);
	///\brief Get the size of the array
	int32_t size() const;
	///\brief Remove a range of elements from the array
	void removeRange(int32_t from, int32_t to);
};
}}

namespace net {
namespace strus {
namespace api {
/// \brief Fixed size vector of strings
class StringVector :public java::util::AbstractList<std::string>
{
	///\brief Create a String vector with a preallocated size
	///\param[in] size number of elements to preallocate
	StringVector(int32_t size);
};

/// \brief Fixed size vector of normalizer definitions
class NormalizerVector :public java::util::AbstractList<Normalizer>
{
public:
	///\brief Create a Normalizer vector with a preallocated size
	///\param[in] size number of elements to preallocate
	NormalizerVector(int32_t size);
};

/// \brief Fixed size vector of query/document terms
class TermVector :public java::util::AbstractList<Term>
{
public:
	///\brief Create a Term vector with a preallocated size
	///\param[in] size number of elements to preallocate
	TermVector(int32_t size);
};

/// \brief Fixed size vector of a query rank. Represents the result of an information rerieval query
class RankVector :public java::util::AbstractList<Rank>
{
public:
	///\brief Create a Rank (item of a query result) vector with a preallocated size
	///\param[in] size number of elements to preallocate
	RankVector(int32_t size);
};

/// \brief Fixed size vector of a query rank. Represents the result of an information rerieval query
class RankAttributeVector :public java::util::AbstractList<RankAttribute>
{
public:
	///\brief Create a RankAttribute (attribute of a query result) vector with a preallocated size
	///\param[in] size number of elements to preallocate
	RankAttributeVector(int32_t size);
};

/// \brief Fixed size vector of document attributes
class AttributeVector :public java::util::AbstractList<Attribute>
{
public:
	///\brief Create a document attribute vector with a preallocated size
	///\param[in] size number of elements to preallocate
	AttributeVector(int32_t size);
};

/// \brief Fixed size vector of document meta data elements
class MetaDataVector :public java::util::AbstractList<MetaData>
{
public:
	///\brief Create a document meta data vector with a preallocated size
	///\param[in] size number of elements to preallocate
	MetaDataVector(int32_t size);
};
}}}

#elif defined DOXYGEN_PHP
/// \brief A single PHP string or an array  of strings
struct StringVector;

/// \brief A structure representing a sequence of normalization functions. Either a PHP string describing a single function by name without parameters or an array of function objects, each function object represented as string for a function name or an array with the first element being the function name and the following elements the parameters passed to the function.
/// \example "lc"                                   :Single function doing lowercase as normalization
/// \example [["stem,"de"]]                         :Single function doing german (de) stemming with snowball as normalization
/// \example ["lc",["stem,"de"],["convdia","de"]]   :Three functions doing in sequence lowercase, german (de) stemming and conversion of diacritical characters as normalization
class NormalizerVector;

/// \brief A structure representing a tokenization function. Either a PHP string describing a single function by name without parameters or an array with the first element being the function name and the following elements the parameters passed to the function.
class Tokenizer;

/// \brief 
#ifdef TOTO_UNFINISHED
class TermVector;
class RankVector;
class RankAttributeVector;
class AttributeVector;
class MetaDataVector;
#endif

#elif defined DOXYGEN_PYTHON
#else
#error Unknown language (swig.hpp)
#endif

